import logging

from stcal.alignment.util import compute_s_region_keyword, wcs_bbox_from_shape

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def create_footprint(wcs, shape=None, center=False):
    """Calculate sky footprint

    Parameters
    ----------
    wcs : `gwcs.WCS`
        The WCS information to get the footprint from

    shape : n-tuple or None
       Shape to use if wcs has no defined shape.

    center : bool
        If True use the center of the pixel, otherwise use the corner.

    Returns
    -------
    footprint : `numpy.ndarray`
        The footprint.
    """
    bbox = wcs.bounding_box

    if bbox is None:
        bbox = wcs_bbox_from_shape(shape)

    # footprint is an array of shape (2, 4) - i.e. 4 values for RA and 4 values for
    # Dec - as we are interested only in the footprint on the sky
    footprint = wcs.footprint(bbox, center=center, axis_type="spatial").T
    # take only imaging footprint
    footprint = footprint[:2, :]

    # Make sure RA values are all positive
    negative_ind = footprint[0] < 0
    if negative_ind.any():
        footprint[0][negative_ind] = 360 + footprint[0][negative_ind]

    footprint = footprint.T
    return footprint


def add_s_region(model):
    """
    Calculate the detector's footprint using ``WCS.footprint`` and save it in the
    ``S_REGION`` keyword

    Parameters
    ----------
    model : `~roman_datamodels.datamodels.ImageModel`
        The data model for processing

    Returns
    -------
    A formatted string representing the detector's footprint
    """
    _update_s_region_keyword(
        model, create_footprint(model.meta.wcs, shape=model.shape, center=False)
    )


def _update_s_region_keyword(model, footprint):
    s_region = compute_s_region_keyword(footprint)
    log.info(f"S_REGION VALUES: {s_region}")
    if "nan" in s_region:
        # do not update s_region if there are NaNs.
        log.info("There are NaNs in s_region, S_REGION not updated.")
    else:
        model.meta.wcsinfo.s_region = s_region
        log.info(f"Update S_REGION to {model.meta.wcsinfo.s_region}")



"""Project default for pytest"""

import inspect
import json
import logging
import os
import sys
from io import StringIO

import pytest
from astropy import coordinates as coord
from astropy import units as u
from astropy.modeling.models import Shift
from gwcs import coordinate_frames as cf
from gwcs import wcs
from roman_datamodels import datamodels as rdm
from roman_datamodels import stnode

from romancal.assign_wcs import pointing
from romancal.assign_wcs.utils import add_s_region

collect_ignore = ["lib/dqflags.py"]


@pytest.fixture
def slow(request):
    """Setup slow fixture for tests to identify if --slow
    has been specified
    """
    return request.config.getoption("--slow")


@pytest.fixture(scope="session")
def dms_logger():
    """Set up a 'DMS' logger for use in tests.

    The primary use is to report DMS requirement log
    messages to stderr but can also be used for general
    stderr logging in tests.
    """
    logger = logging.getLogger("DMS")
    # Don't propagate to root logger to avoid double reporting
    # during stpipe API calls (like Step.call).
    logger.propagate = False
    handler = logging.StreamHandler(sys.stdout)
    handler.setFormatter(
        logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    )
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    return logger


@pytest.fixture(scope="function")
def function_jail(tmp_path):
    """Perform test in a pristine temporary working directory."""
    old_dir = os.getcwd()
    os.chdir(tmp_path)
    try:
        yield str(tmp_path)
    finally:
        os.chdir(old_dir)


@pytest.fixture(scope="module")
def module_jail(request, tmp_path_factory):
    """Run test in a pristine temporary working directory, scoped to module.

    This fixture is the same as _jail in ci_watson, but scoped to module
    instead of function.  This allows a fixture using it to produce files in a
    temporary directory, and then have the tests access them.
    """
    old_dir = os.getcwd()
    path = request.module.__name__.split(".")[-1]
    if request._parent_request.fixturename is not None:
        path = path + "_" + request._parent_request.fixturename
    newpath = tmp_path_factory.mktemp(path)
    os.chdir(str(newpath))
    yield newpath
    os.chdir(old_dir)


@pytest.hookimpl(trylast=True)
def pytest_configure(config):
    terminal_reporter = config.pluginmanager.getplugin("terminalreporter")
    config.pluginmanager.register(
        TestDescriptionPlugin(terminal_reporter), "testdescription"
    )


class TestDescriptionPlugin:
    """Pytest plugin to print the test docstring when `pytest -vv` is used.

    This plug-in was added to support JWST instrument team testing and
    reporting for the JWST calibration pipeline.
    """

    def __init__(self, terminal_reporter):
        self.terminal_reporter = terminal_reporter
        self.desc = None

    def pytest_runtest_protocol(self, item):
        try:
            # Get the docstring for the test
            self.desc = inspect.getdoc(item.obj)
        except AttributeError:
            self.desc = None

    @pytest.hookimpl(hookwrapper=True, tryfirst=True)
    def pytest_runtest_logstart(self, nodeid, location):
        # When run as `pytest` or `pytest -v`, no change in behavior
        if self.terminal_reporter.verbosity <= 1:
            yield
        # When run as `pytest -vv`, `pytest -vvv`, etc, print the test docstring
        else:
            self.terminal_reporter.write("\n")
            yield
            if self.desc:
                self.terminal_reporter.write(f"\n{self.desc} ")


@pytest.fixture(scope="function")
def create_mock_asn_file():
    def _create_asn_file(tmp_path: str, members_mapping: dict | None = None) -> str:
        """
        Create a mock association file with the provided members mapping.

        Parameters
        ----------
        tmp_path : str
            Path to the temporary directory to store the association file.
        members_mapping : list, optional
            Mapping of members to include in the association file (default: None).

        Returns
        -------
        str
            Path to the created association file.
        """
        asn_content = """
            {
                "asn_type": "None",
                "asn_rule": "DMS_ELPP_Base",
                "version_id": null,
                "code_version": "0.9.1.dev28+ge987cc9.d20230106",
                "degraded_status": "No known degraded exposures in association.",
                "program": "noprogram",
                "constraints": "No constraints",
                "asn_id": "a3001",
                "target": "none",
                "asn_pool": "test_pool_name",
                "products": [
                    {
                        "name": "files.asdf",
                        "members": [
                            {
                                "expname": "img_1.asdf",
                                "exptype": "science"
                            },
                            {
                                "expname": "img_2.asdf",
                                "exptype": "science"
                            }
                        ]
                    }
                ]
            }
        """
        if members_mapping:
            asn_dict = json.loads(asn_content)
            asn_dict["products"][0]["members"] = []
            for x in members_mapping:
                asn_dict["products"][0]["members"].append(x)
            asn_content = json.dumps(asn_dict)

        asn_file_path = str(tmp_path / "sample_asn.json")
        asn_file = StringIO()
        asn_file.write(asn_content)
        with open(asn_file_path, mode="w") as f:
            print(asn_file.getvalue(), file=f)

        return asn_file_path

    return _create_asn_file


def _create_wcs(input_dm, shift_1=0, shift_2=0):
    """
    Create a basic WCS object (with optional shift) and
    append it to the input_dm.meta attribute.

    The WCS object will have a pipeline with the required
    steps to validate against the TweakReg pipeline.

    Parameters
    ----------
    input_dm : roman_datamodels.ImageModel
        A Roman image datamodel.
    shift_1 : int, optional
        Shift to be applied in the direction of the first axis, by default 0
    shift_2 : int, optional
        Shift to be applied in the direction of the second axis, by default 0
    """

    shape = input_dm.data.shape

    # create necessary transformations
    distortion = Shift(-shift_1) & Shift(-shift_2)
    tel2sky = pointing.v23tosky(input_dm)

    # create required frames
    detector = cf.Frame2D(name="detector", axes_order=(0, 1), unit=(u.pix, u.pix))
    v2v3 = cf.Frame2D(
        name="v2v3",
        axes_order=(0, 1),
        axes_names=("v2", "v3"),
        unit=(u.arcsec, u.arcsec),
    )
    world = cf.CelestialFrame(reference_frame=coord.ICRS(), name="world")

    # create pipeline
    pipeline = [
        wcs.Step(detector, distortion),
        wcs.Step(v2v3, tel2sky),
        wcs.Step(world, None),
    ]

    wcs_obj = wcs.WCS(pipeline)
    wcs_obj.bounding_box = ((-0.5, shape[-2] + 0.5), (-0.5, shape[-1] + 0.5))

    input_dm.meta["wcs"] = wcs_obj

    add_s_region(input_dm)


def _base_image(shift_1=0, shift_2=0):
    l2 = rdm.ImageModel.create_fake_data(shape=(100, 100))
    l2.meta.filename = "none"
    l2.meta.cal_logs = stnode.CalLogs.create_fake_data()
    l2.meta.cal_step = stnode.L2CalStep.create_fake_data()
    l2.meta.background = stnode.SkyBackground.create_fake_data()
    l2.var_flat = l2.var_rnoise.copy()
    _create_wcs(l2)
    l2.meta.wcsinfo.vparity = -1
    return l2


@pytest.fixture
def base_image():
    """
    Create a base image with a realistic WCSInfo and a WCS.

    Notes
    -----
    The size of the image needs to be relatively large in order for
    the source catalog step to find a reasonable number of sources in the image.

    shift_1 and shift_2 (units in pixel) are used to shift the WCS projection plane.
    """

    return _base_image



"""
Test suite for engdb_tools
"""

import logging
import os

import pytest
from astropy.time import Time

from romancal.lib.engdb import engdb_tools

# Configure logging
log = logging.getLogger(__name__)

GOOD_MNEMONIC = "OPE_SCF_DIR"
GOOD_STARTTIME = "2027-02-23T01:00:00"
GOOD_ENDTIME = "2027-02-23T01:01:00"

SHORT_STARTTIME = "2027-02-23T01:00:30"

BAD_MNEMONIC = "No_Such_MNEMONIC"
NODATA_STARTIME = "2014-01-01"
NODATA_ENDTIME = "2014-01-02"


def test_environmental_bad(monkeypatch):
    alternate = "https://google.com/"
    did_except = False
    monkeypatch.setenv("ENG_BASE_URL", alternate)
    try:
        engdb = engdb_tools.engdb_service("mast")
    except Exception:
        did_except = True
    assert did_except, f"DB connection falsely created for {engdb.base_url}"


def test_basic(engdb):
    assert engdb._get_records(GOOD_MNEMONIC, GOOD_STARTTIME, GOOD_ENDTIME)


def test_bad_service():
    with pytest.raises(RuntimeError):
        engdb_tools.engdb_service("junk_service")


def test_values(engdb):
    records = engdb._get_records(GOOD_MNEMONIC, SHORT_STARTTIME, SHORT_STARTTIME)
    assert len(records) == 2
    values = engdb.get_values(GOOD_MNEMONIC, GOOD_STARTTIME, SHORT_STARTTIME)
    assert len(values) == 29
    assert values[0] == "SCFA"


def test_values_with_bracket(engdb):
    records = engdb._get_records(GOOD_MNEMONIC, SHORT_STARTTIME, SHORT_STARTTIME)
    assert len(records) == 2
    values = engdb.get_values(
        GOOD_MNEMONIC, SHORT_STARTTIME, SHORT_STARTTIME, include_bracket_values=True
    )
    assert len(values) == 2
    assert values[1] == "SCFA"


def test_values_with_time(engdb):
    values = engdb.get_values(
        GOOD_MNEMONIC, GOOD_STARTTIME, SHORT_STARTTIME, include_obstime=True
    )
    assert len(values) >= 1
    assert isinstance(values[0], tuple)
    assert isinstance(values[0].obstime, Time)


def test_novalues(engdb):
    values = engdb.get_values(GOOD_MNEMONIC, NODATA_STARTIME, NODATA_ENDTIME)
    assert len(values) == 0


def test_unzip(engdb):
    """Test forunzipped versions of content"""
    values = engdb.get_values(
        GOOD_MNEMONIC,
        SHORT_STARTTIME,
        SHORT_STARTTIME,
        include_obstime=True,
        zip_results=False,
    )
    assert isinstance(values, tuple)
    assert len(values.obstime) == len(values.value)


# ########
# Fixtures
# ########
@pytest.fixture(
    scope="module", params=[name for name in engdb_tools.AVAILABLE_SERVICES]
)
def engdb(request):
    """Setup a service"""
    service = request.param
    args = {
        "mast": {},
        "edp": {
            "environment": os.environ.get("EDP_ENVIRONMENT", "test"),
            "path_to_cc": os.environ.get("PATH_TO_CC", None),
        },
    }
    try:
        engdb = engdb_tools.engdb_service(service, **args[service])
    except RuntimeError as exception:
        pytest.skip(f"Engineering database unvailable: {exception}.")
    yield engdb



import logging

import asdf

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)


def save_median(median_data, median_wcs, make_output_path):
    _save_intermediate_output(
        median_data,
        median_wcs,
        make_output_path("drizzled_median.asdf", suffix="median"),
    )


def save_drizzled(drizzled_model, make_output_path):
    input_path = drizzled_model.meta.get("filename", "none").replace("_outlier_", "_")
    _save_intermediate_output(
        drizzled_model.data,
        drizzled_model.meta.wcs,
        make_output_path(input_path, suffix="outlier_coadd"),
    )


def _save_intermediate_output(data, wcs, output_path):
    asdf.AsdfFile({"data": data, "wcs": wcs}).write_to(output_path)
    log.info(f"Saved intermediate product to {output_path}")



"""
This module determines which sky cells overlap with the given image.

Currently this assumes that the sky projected borders of all calibrated L2 images are straight.
"""

import logging
from functools import cached_property

import numpy as np
import spherical_geometry.great_circle_arc as sga
import spherical_geometry.polygon as sgp
import spherical_geometry.vector as sgv
from gwcs import WCS
from numpy.typing import NDArray

import romancal.skycell.skymap as sc

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

__all__ = ["ImageFootprint", "find_skycell_matches"]


class ImageFootprint:
    """abstraction of an image footprint"""

    _radec_corners: NDArray[float]

    def __init__(self, radec_vertices: list[tuple[float, float]]):
        """
        Parameters
        ----------
        radec_vertices: list[tuple[float, float]]
            vertices (usually the corners) of the image in right ascension and declination
        """
        self._radec_vertices = np.array(radec_vertices)

    @classmethod
    def from_wcs(
        cls,
        wcs: WCS,
        extra_vertices_per_edge: int = 0,
    ) -> "ImageFootprint":
        """create an image footprint from a GWCS object (and image shape, if no bounding box is present)

        Parameters
        ----------
        wcs: WCS :
            WCS object
        extra_vertices_per_edge: int :
            extra vertices to create on each edge to capture distortion (Default value = 0)

        Returns
        -------
        image footprint object
        """

        if not hasattr(wcs, "bounding_box") or wcs.bounding_box is None:
            raise ValueError(
                "Cannot infer image footprint from WCS without a bounding box."
            )

        array_shape = (
            wcs.array_shape
            if hasattr(wcs, "array_shape") and wcs.array_shape is not None
            else tuple(
                wcs.bounding_box[index][1] - wcs.bounding_box[index][0]
                for index in range(len(wcs.bounding_box))
            )
        )
        if extra_vertices_per_edge <= 0:
            vertex_points = wcs.footprint(center=False)
        else:
            # constrain number of vertices to the maximum number of pixels on an edge, excluding the corners
            if extra_vertices_per_edge > max(array_shape) - 2:
                extra_vertices_per_edge = max(array_shape) - 2

            # build a list of pixel indices that represent equally-spaced edge vertices
            vertices_per_edge = 2 + extra_vertices_per_edge
            origin_indices = np.zeros(vertices_per_edge - 1) - 0.5
            x_end_indices = array_shape[0] - origin_indices
            y_end_indices = array_shape[1] - origin_indices
            vertices_x = np.linspace(
                0, array_shape[0], num=vertices_per_edge - 1, endpoint=False
            )
            vertices_y = np.linspace(
                0, array_shape[1], num=vertices_per_edge - 1, endpoint=False
            )
            edge_indices = np.concatenate(
                [
                    # north edge
                    np.stack([origin_indices, vertices_y], axis=1),
                    # east edge
                    np.stack([vertices_x, y_end_indices], axis=1),
                    # south edge
                    np.stack([x_end_indices, y_end_indices - vertices_y], axis=1),
                    # west edge
                    np.stack([x_end_indices - vertices_x, origin_indices], axis=1),
                ],
                axis=0,
            )

            # query the WCS for pixel indices at the edges
            vertex_points = np.stack(
                wcs(*edge_indices.T, with_bounding_box=False),
                axis=1,
            )

        return cls(vertex_points)

    @property
    def radec_corners(self) -> NDArray:
        """vertices in right ascension and declination in counterclockwise order"""
        return self._radec_vertices

    @cached_property
    def radec_center(self) -> tuple[float, float]:
        """center point in right ascension and declination"""
        return sgv.vector_to_lonlat(*self.vectorpoint_center)

    @cached_property
    def vectorpoint_vertices(self) -> NDArray[float]:
        """vertices in 3D Cartesian space on the unit sphere"""
        return sgv.normalize_vector(
            np.stack(sgv.lonlat_to_vector(*np.array(self.radec_corners).T), axis=1)
        )

    @cached_property
    def vectorpoint_center(self) -> tuple[float, float, float]:
        """center in 3D Cartesian space on the unit sphere"""
        return sgv.normalize_vector(np.mean(self.vectorpoint_vertices, axis=0))

    @cached_property
    def length(self) -> float:
        """diagonal length of the rectangular footprint"""
        # assume equally-spaced points around the perimeter
        # NOTE: this will produce an incorrect value with no error if the points are not equally spaced
        half_index_length = round(len(self.vectorpoint_vertices) / 2)
        return max(
            sga.length(
                self.vectorpoint_vertices[index],
                self.vectorpoint_vertices[index + half_index_length],
            )
            for index in range(len(self.vectorpoint_vertices) - half_index_length - 1)
        )

    @cached_property
    def circumference(self) -> float:
        """circumference of the rectangular footprint"""
        return sum(
            sga.length(
                self.vectorpoint_vertices[index], self.vectorpoint_vertices[index + 1]
            )
            for index in range(-1, len(self.vectorpoint_vertices) - 1)
        )

    @cached_property
    def polygon(self) -> sgp.SingleSphericalPolygon:
        """spherical polygon representing this image footprint"""
        return sgp.SingleSphericalPolygon(
            points=self.vectorpoint_vertices,
            inside=self.vectorpoint_center,
        )

    @cached_property
    def area(self) -> float:
        """area of this footprint on the sphere in degrees squared"""
        return self.polygon.area()

    @cached_property
    def possibly_intersecting_projregions(self) -> int:
        """number of possibly intersecting projection regions"""
        if self.area > sc.SkyCell.area:
            return (
                # the number of times the smallest projection region could fit in the image footprint
                np.ceil(self.area / sc.ProjectionRegion.MIN_AREA)
                # plus multiplier for partial intersections on the perimeter
                * 8
            )
        else:
            # 4 foundational intersections
            return 4

    @cached_property
    def possibly_intersecting_skycells(self) -> int:
        """number of possibly intersecting skycells"""
        if self.polygon.area() > sc.SkyCell.area:
            return (
                # number of times a skycell could fit in the image footprint
                np.ceil(self.polygon.area() / sc.SkyCell.area)
                # plus multiplier for partial intersections on the perimeter
                * 8
            )
        else:
            # 4 foundational intersections
            return 4

    @cached_property
    def possible_intersecting_projregion_distance(self) -> int:
        """maximum possible distance to the center of an intersecting projection region"""
        return (self.length + sc.ProjectionRegion.MAX_LENGTH) / 2.0

    @cached_property
    def possible_intersecting_skycell_distance(self) -> int:
        """maximum possible distance to the center of an intersecting sky cell"""
        return (self.length + sc.SkyCell.length) / 2.0

    def __str__(self) -> str:
        return f"footprint {self.radec_corners}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.radec_corners!r})"


def find_skycell_matches(
    image_corners: list[tuple[float, float]] | NDArray[float] | WCS,
    skymap: sc.SkyMap = None,
) -> list[int]:
    """Find sky cells overlapping the provided image footprint

    Parameters
    ----------
    image_corners : list | np.ndarray | WCS :
        Either a squence of 4 (ra, dec) pairs, or
        equivalent 2-d numpy array, or a GWCS instance.
        A GWCS instance must have `.bounding_box` or `.pixel_shape` attribute defined.
    skymap : sc.SkyMap :
        sky map instance; defaults to global SKYMAP (Default value = None)

    Returns
    -------
    A sequence of the indices of all sky cells that overlap the supplied image
    (in the referenced sky map). The indices may be used to obtain all
    necessary information about the sky cells, by calling `romancal.skycell.skymap.SkyCell(index)`.
    """

    if isinstance(image_corners, WCS):
        footprint = ImageFootprint.from_wcs(image_corners, extra_vertices_per_edge=3)
    else:
        footprint = ImageFootprint(image_corners)

    if skymap is None:
        skymap = sc.SKYMAP

    intersecting_skycell_indices = []

    # query the global k-d tree of projection regions for possible intersection candidates in (normalized) 3D space
    nearby_projregion_indices = np.array(
        skymap.projection_regions_kdtree.query(
            footprint.vectorpoint_center,
            # NOTE: `k` can be `footprint.possibly_intersecting_projregions` if we assume non-degenerate tesselation
            k=100,
            distance_upper_bound=footprint.possible_intersecting_projregion_distance
            * 1.1,
        )[1]
    )
    nearby_projregion_indices = nearby_projregion_indices[
        np.where(nearby_projregion_indices != len(skymap.model.projection_regions))
    ]

    for projregion_index in nearby_projregion_indices:
        projregion = sc.ProjectionRegion(projregion_index)
        if footprint.polygon.intersects_poly(projregion.polygon):
            # query the LOCAL k-d tree of skycells for possible intersection candidates in (normalized) 3D space
            projregion_nearby_skycell_indices = np.array(
                projregion.skycells_kdtree.query(
                    footprint.vectorpoint_center,
                    # NOTE: `k` can be `footprint.possibly_intersecting_skycells` if we assume non-degenerate tesselation
                    k=100,
                    distance_upper_bound=footprint.possible_intersecting_skycell_distance
                    * 1.1,
                )[1]
            )
            projregion_nearby_skycell_indices = np.array(
                projregion_nearby_skycell_indices[
                    np.where(
                        projregion_nearby_skycell_indices != len(projregion.skycells)
                    )
                ]
            )

            # convert to absolute indices over the full skycell table
            projregion_nearby_skycell_indices += projregion.data["skycell_start"]

            # find polygons that intersect the image footprint
            for skycell_index in projregion_nearby_skycell_indices:
                # print(nearby_skycell_index)
                skycell = sc.SkyCell(skycell_index)
                if footprint.polygon.intersects_poly(skycell.polygon):
                    # print(f"candidate {nearby_skycell_index} intersects")
                    intersecting_skycell_indices.append(skycell_index)

    return np.array(intersecting_skycell_indices)



"""Resource tracking regtest utilities

Can be used within module-scoped fixtures (often used to
run Steps or Pipelines) or within tests.

For uses where the resource usage occurs within a test:

>>>
>> def test_long_step(resource_tracker, request):
>>     with resource_tracker.track(log=request):
>>         # something that takes memory and time
>>         pass

For a module-scoped fixture the resource tracking can
be performed in the fixture but the logging/reporting of
the resource usage must occur during a test:

>>>
>> @pytest.fixture(scope="module")
>> def resource_tracker():
>>     return ResourceTracker()
>>
>> @pytest.fixture()
>> def log_tracked_resources(resource_tracker, request):
>>     def callback():
>>         resource_tracker.log(request)
>>
>>     yield callback
>>
>> @pytest.fixture
>> def my_long_fixture(resource_tracker):
>>     with resource_tracker.track():
>>         # something that takes memory and time
>>         pass
>>
>> def test_log_tracked_resources(log_tracked_resources, my_long_fixture):
>>     log_tracked_resources()

Use of the module-scoped fixture has fixture-reuse
considerations similar to the ``rtdata_module`` fixture. Having
more than one module scoped fixture that uses ``resource_tracker``
per module is discouraged (as both will use the same ``ResourceTracker``
instance). Parameterization of a fixture using ``resource_tracker``
is supported (same as ``rtdata_module``).
"""

import time
import tracemalloc
from contextlib import ExitStack, contextmanager


class TrackRuntime:
    """Runtime tracker context."""

    def __enter__(self):
        self._t0 = time.monotonic()

    def __exit__(self, exc_type, exc_value, traceback):
        self.value = time.monotonic() - self._t0

    def log(self):
        return ("tracked-time", self.value)


class TrackPeakMemory:
    """Peak memory tracker context."""

    def __enter__(self):
        tracemalloc.start()

    def __exit__(self, exc_type, exc_value, traceback):
        _, self.value = tracemalloc.get_traced_memory()
        tracemalloc.stop()

    def log(self):
        return ("tracked-peakmem", self.value)


class ResourceTracker:
    """Track resources used during track context."""

    def __init__(self):
        self._trackers = [TrackPeakMemory(), TrackRuntime()]

    def log(self, request):
        """Log tracked resource usage to the pytest request user properties.

        Parameters
        ----------
        request : pytest.FixtureRequest
            Must be a function-scoped pytest request fixture result.
        """
        request.node.user_properties.extend(t.log() for t in self._trackers)

    @contextmanager
    def track(self, log=None):
        """Context during which resources are tracked.

        Parameters
        ----------
        log : pytest.FixtureRequest, optional
            If provided, log the usage to the provided request fixture result.
        """
        try:
            with ExitStack() as stack:
                [stack.enter_context(t) for t in self._trackers]
                yield self
        finally:
            if log:
                self.log(log)



import logging

import numpy as np
from astropy.modeling.models import Identity, RotationSequence3D, Scale, Shift
from gwcs.geometry import CartesianToSpherical, SphericalToCartesian

log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


def v23tosky(input_model, wrap_v2_at=180, wrap_lon_at=360):
    """Create the transform from telescope to sky.

    The transform is defined with a reference point in a Frame
    associated tih the telescope (V2, V3) in arcsec, the corresponding
    reference poiont on sky (RA_REF, DEC_REF) in deg, and the position angle
    at the center of the aperture, ROLL_REF in deg.

    Parameters
    ----------
    input_model : `roman_datamodels.WfiImage`
        Roman imaging exposure data model.
    wrap_v2_at : float
        At what angle to wrap V2. [deg]
    wrap_lon_at : float
        At what angle to wrap logitude. [deg]

    Returns
    -------
    model : `astropy.modeling.Model`
        The transform from V2,V3 to sky.
    """
    v2_ref = input_model.meta.wcsinfo.v2_ref / 3600
    v3_ref = input_model.meta.wcsinfo.v3_ref / 3600
    roll_ref = input_model.meta.wcsinfo.roll_ref
    ra_ref = input_model.meta.wcsinfo.ra_ref
    dec_ref = input_model.meta.wcsinfo.dec_ref

    angles = np.array([v2_ref, -v3_ref, roll_ref, dec_ref, -ra_ref])
    axes = "zyxyz"
    rot = RotationSequence3D(angles, axes_order=axes)

    # The sky rotation expects values in deg.
    # This should be removed when models work with quantities.
    model = (
        (Scale(1 / 3600) & Scale(1 / 3600))
        | SphericalToCartesian(wrap_lon_at=wrap_v2_at)
        | rot
        | CartesianToSpherical(wrap_lon_at=wrap_lon_at)
    )
    model.name = "v23tosky"
    return model


def dva_corr_model(va_scale, v2_ref, v3_ref):
    """
    Create transformation that accounts for differential velocity aberration
    (scale).

    Parameters
    ----------
    va_scale : float, None
        Ratio of the apparent plate scale to the true plate scale. When
        ``va_scale`` is `None`, it is assumed to be identical to ``1`` and
        an ``astropy.modeling.models.Identity`` model will be returned.

    v2_ref : float, None
        Telescope ``v2`` coordinate of the reference point in ``arcsec``. When
        ``v2_ref`` is `None`, it is assumed to be identical to ``0``.

    v3_ref : float, None
        Telescope ``v3`` coordinate of the reference point in ``arcsec``. When
        ``v3_ref`` is `None`, it is assumed to be identical to ``0``.

    Returns
    -------
    va_corr : astropy.modeling.CompoundModel, astropy.modeling.models.Identity
        A 2D compound model that corrects DVA. If ``va_scale`` is `None` or 1
        then `astropy.modeling.models.Identity` will be returned.

    """
    if va_scale is None or va_scale == 1:
        return Identity(2)

    if va_scale <= 0:
        log.warning("Given velocity aberration scale %s", va_scale)
        log.warning(
            "Velocity aberration scale must be a positive number. Setting to 1.0"
        )
        va_scale = 1.0

    va_corr = Scale(va_scale, name="dva_scale_v2") & Scale(
        va_scale, name="dva_scale_v3"
    )

    if v2_ref is None:
        v2_ref = 0

    if v3_ref is None:
        v3_ref = 0

    if v2_ref == 0 and v3_ref == 0:
        return va_corr

    # NOTE: it is assumed that v2, v3 angles and va scale are small enough
    # so that for expected scale factors the issue of angle wrapping
    # (180 degrees) can be neglected.
    v2_shift = (1 - va_scale) * v2_ref
    v3_shift = (1 - va_scale) * v3_ref

    va_corr |= Shift(v2_shift, name="dva_v2_shift") & Shift(
        v3_shift, name="dva_v3_shift"
    )
    va_corr.name = "DVA_Correction"
    return va_corr



from collections import defaultdict
from typing import ClassVar

import numpy as np
from asdf.lazy_nodes import AsdfDictNode, AsdfListNode
from asdf.tags.core.ndarray import NDArrayType
from astropy.table import Table
from astropy.time import Time
from roman_datamodels import datamodels, stnode


class MissingCellType:
    def __str__(self):
        return "MISSING_CELL"


MISSING_CELL = MissingCellType()


class TableBuilder:
    def __init__(self):
        self._columns = {}
        self._flagged_columns = set()

    def _get_column(self, name, row_index):
        if name not in self._columns:
            if row_index:
                self._flagged_columns.add(name)
            self._columns[name] = [MISSING_CELL] * row_index
        return self._columns[name]

    def _sanitize_value(self, value):
        if isinstance(value, list | dict | AsdfDictNode | AsdfListNode):
            return str(value)
        return value

    def add_row(self, data_dict, row_index):
        updated = set()
        for key, value in data_dict.items():
            svalue = self._sanitize_value(value)
            if svalue is None:
                self._flagged_columns.add(key)
            self._get_column(key, row_index).append(svalue)
            updated.add(key)
        for key in self._columns.keys() - updated:
            self._flagged_columns.add(key)
            self._columns[key].append(MISSING_CELL)

    def to_table(self):
        array_columns = {}
        for name, values in self._columns.items():
            if name not in self._flagged_columns:
                array_columns[name] = values
                continue
            # this column has either a None or MISSING_CELL which
            # has to be filled in
            arr = np.array([v for v in values if v not in (None, MISSING_CELL)])

            # if we have no valid values, return all "None"
            if not arr.size:
                array_columns[name] = ["None"] * len(values)
                continue

            # if we have a float or int use 'nan' for missing/None values
            # this will convert int columns to floats
            if arr.dtype.kind in ("f", "i"):
                array_columns[name] = [
                    np.nan if v in (MISSING_CELL, None) else v for v in values
                ]
                continue

            # if all else fails, convert everything to a string
            array_columns[name] = [str(v) for v in values]

        return Table(array_columns)


class MetaBlender:
    _meta_blend_paths: ClassVar = {
        "observation.execution_plan": None,
        "observation.exposure": None,
        "observation.observation": None,
        "observation.pass": None,
        "observation.program": None,
        "observation.segment": None,
        "observation.visit": None,
        "program.title": None,
        "program.investigator_name": None,
        "program.category": None,
        "program.subcategory": None,
        "program.science_category": None,
        "instrument.optical_element": None,
        "cal_step.flux": "INCOMPLETE",
        "cal_step.outlier_detection": "INCOMPLETE",
        "cal_step.skymatch": "INCOMPLETE",
    }

    def __init__(self):
        self._tables = defaultdict(TableBuilder)
        self._n_rows = 0

    def _blend_first(self, model):
        # make a blank mosic metdata node
        self._model = datamodels.MosaicModel.create_minimal()

        # FIXME assuming everything is a prompt coadd
        self._model.meta.product_type = stnode.ProductType("p_visit_coadd")

        self._meta = self._model.meta

        self._model["individual_image_cal_logs"] = []

        # for computing mean start time
        self._start_times = []

        # grab start/end times, blended with all other models below
        self._meta.coadd_info.time_first = model.meta.exposure.start_time
        self._meta.coadd_info.time_last = model.meta.exposure.end_time

        # copy over metadata from first model
        for path, default in self._meta_blend_paths.items():
            dst = self._meta
            src = model.meta
            *sub_path, key = path.split(".")
            for k in sub_path:
                src = src.get(k, {})
                dst = dst[k]
            dst[key] = src.get(key, default)

    def _update_tables(self, meta):
        basic_data = {}
        for key, value in meta.to_flat_dict().items():
            if key in {"wcs", "cal_logs"}:
                continue

            if isinstance(value, stnode.DNode):
                self._tables[key].add_row(value, self._n_rows)
            elif not isinstance(value, dict | NDArrayType | Table | AsdfDictNode):
                basic_data[key] = value
        if basic_data:
            self._tables["basic"].add_row(basic_data, self._n_rows)
        self._n_rows += 1

    def blend(self, model):
        if not hasattr(self, "_model"):
            self._blend_first(model)
        else:
            # for non-first only blending
            self._meta.coadd_info.time_first = min(
                self._meta.coadd_info.time_first, model.meta.exposure.start_time
            )
            self._meta.coadd_info.time_last = max(
                self._meta.coadd_info.time_last, model.meta.exposure.end_time
            )

            # blend
            for path, default in self._meta_blend_paths.items():
                dst = self._meta
                src = model.meta
                *sub_path, key = path.split(".")
                for k in sub_path:
                    src = src.get(k, {})
                    dst = dst[k]
                if key not in src:
                    continue
                if dst[key] != src[key]:
                    dst[key] = default

        self._start_times.append(model.meta.exposure.start_time)

        if cal_logs := model.meta.get("cal_logs"):
            self._model["individual_image_cal_logs"].append(cal_logs)

        self._meta.resample.members.append(model.meta.filename)

        self._update_tables(model.meta)

    def finalize(self):
        self._meta.coadd_info.time_mean = Time(self._start_times).mean()
        if all(
            getattr(self._meta.observation, key) is not None
            for key in ("execution_plan", "pass", "segment", "observation", "visit")
        ):
            self._meta.observation.exposure_grouping = "v{execution_plan:02d}{pass:03d}{segment:03d}{observation:03d}{visit:03d}".format(
                **self._meta.observation
            )
        else:
            self._meta.observation.exposure_grouping = None
        self._meta.individual_image_meta = {}
        for table_name, builder in self._tables.items():
            self._meta.individual_image_meta[table_name] = builder.to_table()
        return self._model



import logging

import numpy as np
from roman_datamodels import datamodels, dqflags
from stcal.resample import Resample

import romancal.skycell.skymap as sc

from .exptime_resampler import ExptimeResampler
from .l3_wcs import assign_l3_wcs
from .meta_blender import MetaBlender
from .resample_utils import make_output_wcs

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

__all__ = ["ResampleData"]


class ResampleData(Resample):
    dq_flag_name_map = dqflags.pixel

    def __init__(
        self,
        input_models,
        output_wcs,
        pixfrac,
        kernel,
        fillval,
        weight_type,
        good_bits,
        enable_ctx,
        enable_var,
        compute_err,
        compute_exptime,
        blend_meta,
        resample_on_skycell,
        wcs_kwargs=None,
        variance_array_names=None,
    ):
        """
        See the base class `stcal.resample.resample.Resample` for more details.

        This is primarily used by `romancal.resample.ResampleStep`. See the step
        spec for parameter defaults and more information.

        Parameters
        ----------
        input_models : ModelLibrary
            The library of models to resample.

        output_wcs : dict, None
            Output wcs to resample to. If None one will be computed from
            input_models.

        pixfrac : float
            The fraction of a pixel that the pixel flux is confined to.

        kernel : str
            The name of the resampling kernel.

        fillval : float, None, str
            The value of output pixels that did not have contributions from
            input images' pixels.

        weight_type : str
            The weighting type for adding data.

        good_bits : int, str, None
            The bit mask of pixels to use for resampling.

        enable_ctx : bool
            If `True` include a context array in the resampled output.

        enable_var : bool
            If `True` resample variance arrays.

        compute_err : {"from_var", "driz_err"}, None
            If `None` do not compute an output model error array. For
            other options see `stcal.resample.resample.Resample`.

        compute_exptime : bool
            If `True` resample input exposure times to compute output
            exposure time information.

        blend_meta : bool
            If `True` include blended input model metadata in the output
            model.

        resample_on_skycell : bool
            If `True` and the association contains skycell information
            use the skycell wcs.

        wcs_kwargs : dict, None
            A dictionary of custom WCS parameters used to define an
            output WCS from input models' outlines. This argument is ignored
            when ``output_wcs`` is specified. See
            `romancal.resample.resample_utils.make_output_wcs`
            for supported options.

        variance_array_names : list, None
            Variance arrays to resample.  If None, use stcal default.
        """
        # fillval indef doesn't define a starting value
        # since we're not resampling onto an existing array
        # we overwrite this to 0 (to match the old behavior)
        if fillval.lower() == "indef":
            fillval = 0
        self.input_models = input_models

        self._compute_exptime = compute_exptime
        self._blend_meta = blend_meta

        if output_wcs is None and resample_on_skycell:
            # first try to retrieve a sky cell name from the association
            try:
                skycell = sc.SkyCell.from_asn(self.input_models.asn)

                log.info(f"Skycell record: {skycell.data}")

                log.info(
                    f"Creating skycell image at ra: {skycell.radec_center[0]}  dec {skycell.radec_center[1]}",
                )
                log.info("Resampling to skycell wcs")
                output_wcs = {"wcs": skycell.wcs}
            except ValueError as err:
                log.warning(f"Unable to compute skycell from input association: {err}")
                log.warning("Computing output wcs from all input wcses")

        if output_wcs is None:
            if wcs_kwargs is None:
                wcs_kwargs = {}
            wcs, ps, ps_ratio = make_output_wcs(input_models, **wcs_kwargs)
            output_wcs = {
                "wcs": wcs,
                "pixel_scale": ps,
                "pixel_scale_ratio": ps_ratio,
            }

        if variance_array_names:
            self.variance_array_names = variance_array_names

        super().__init__(
            output_wcs,
            n_input_models=len(input_models),
            pixfrac=pixfrac,
            kernel=kernel,
            fillval=fillval,
            weight_type=weight_type,
            good_bits=good_bits,
            enable_ctx=enable_ctx,
            enable_var=enable_var,
            compute_err=compute_err,
        )

    @property
    def compute_exptime(self):
        """Indicates if exposure time resampling is enabled."""
        return self._compute_exptime

    @property
    def blend_meta(self):
        """Indicates if metadata blending is enabled."""
        return self._blend_meta

    def _input_model_to_dict(self, model):
        """Convert an input datamodel to a dictionary suitable for the base class"""
        pixel_area = model.meta.photometry.pixel_area
        if pixel_area == -999999:
            pixel_area = None
        exposure_time = model.meta.exposure.exposure_time
        if exposure_time == -999999:
            exposure_time = 1
        if "background" in model.meta:
            level = model.meta.background.level
            subtracted = model.meta.background.subtracted
        else:
            level = 0
            subtracted = True
        return {
            "data": model.data,
            "dq": model.dq,
            "var_rnoise": model.var_rnoise,
            "var_poisson": model.var_poisson,
            "err": model.err,
            "filename": model.meta.filename,
            "wcs": model.meta.wcs,
            "pixelarea_steradians": pixel_area,
            "group_id": self.input_models._model_to_group_id(model),
            "measurement_time": None,  # falls back to exposure_time
            "exposure_time": exposure_time,
            "start_time": model.meta.exposure.start_time,
            "end_time": model.meta.exposure.end_time,
            "duration": None,  # unused
            "level": level,
            "subtracted": subtracted,
            "effective_exposure_time": model.meta.exposure.effective_exposure_time,
            "var_flat": getattr(model, "var_flat", None),
        }

    def _get_intensity_scale(self, model):
        # always provide 1: https://github.com/spacetelescope/romancal/issues/1637
        return 1

    def add_model_hook(self, model, pixmap, iscale, weight_map, xmin, xmax, ymin, ymax):
        if self.compute_exptime:
            if not hasattr(self, "_exptime_resampler"):
                self._exptime_resampler = ExptimeResampler(
                    self.output_wcs,
                    self.output_array_shape,
                    self.good_bits,
                    self.kernel,
                )
            self._exptime_resampler.add_image(model, pixmap, xmin, xmax, ymin, ymax)

    def add_model(self, model):
        model_dict = self._input_model_to_dict(model)
        super().add_model(model_dict)
        if self.blend_meta:
            self._meta_blender.blend(model)

    def finalize(self):
        super().finalize()

        if self.blend_meta:
            output_model = self._meta_blender.finalize()
        else:
            output_model = datamodels.MosaicModel.create_minimal()

        # copy over asn information
        output_model.meta.association.name = self.input_models.asn.get(
            "table_name", "?"
        )

        # resample parameters
        output_model.meta.resample.good_bits = self.good_bits
        output_model.meta.resample.pixel_scale_ratio = self.output_model[
            "pixel_scale_ratio"
        ]
        output_model.meta.resample.pixfrac = self.output_model["pixfrac"]
        output_model.meta.resample.pointings = self.output_model["pointings"]
        output_model.meta.resample.weight_type = self.weight_type

        # every resampling will generate these
        output_model.data = self.output_model["data"]
        output_model.weight = self.output_model["wht"]

        # some things are conditional
        if self.compute_exptime and hasattr(self, "_exptime_resampler"):
            exptime_total = self._exptime_resampler.finalize()
            m = exptime_total > 0
            total_exposure_time = (
                np.mean(exptime_total[m], dtype="f8") if np.any(m) else 0
            )
            max_exposure_time = np.max(exptime_total)
            log.info(
                f"Mean, max exposure times: {total_exposure_time:.1f}, "
                f"{max_exposure_time:.1f}"
            )
            output_model.meta.coadd_info.max_exposure_time = max_exposure_time
            output_model.meta.coadd_info.exposure_time = total_exposure_time

        if self._enable_ctx:
            output_model.context = self.output_model["con"].astype(np.uint32)

        for arr_name in ["err", *self.variance_array_names]:
            if arr_name in self.output_model:
                new_array = self.output_model[arr_name]
                if new_array is not None:
                    setattr(output_model, arr_name, new_array)

        # assign wcs to output model
        assign_l3_wcs(output_model, self.output_wcs)
        output_model.meta.wcsinfo.skycell_name = self.input_models.asn.get(
            "target", "None"
        )

        return output_model

    def reset_arrays(self, n_input_models=None):
        super().reset_arrays(n_input_models)
        if self.blend_meta:
            self._meta_blender = MetaBlender()

    def resample_group(self, indices):
        """
        Resample the models at indices.

        Parameters
        ----------
        indices : iterable
            An iterable that returns the indices of models to resample.

        Returns
        -------
        MosaicModel
            The resampled datamodel
        """
        if self.is_finalized():
            self.reset_arrays(len(indices))

        with self.input_models:
            for index in indices:
                model = self.input_models.borrow(index)
                self.add_model(model)
                self.input_models.shelve(model, index, modify=False)

        return self.finalize()



"""
Plotting utilities for plotting skycells against supplied image

matplotlib dependency is optional.
"""

import numpy as np
import spherical_geometry.vector as sgv
from numpy.typing import NDArray

import romancal.skycell.match as sm
import romancal.skycell.skymap as sc

try:
    from matplotlib import pyplot as plt
    from matplotlib.axis import Axis
except ImportError:
    print("matplotlib is required for this plotting utility")

__all__ = [
    "plot_field",
    "plot_image_footprint_and_skycells",
    "plot_projregion",
    "plot_skycell",
    "veccoords_to_tangent_plane",
]

RAD_TO_ARCSEC = 180.0 / np.pi * 3600.0


def find_intersecting_projregions(
    footprint: sm.ImageFootprint,
    skymap: sc.SkyMap = None,
) -> list[int]:
    """Out of all projection regions, find ones that intersect the given image footprint

    Parameters
    ----------
    footprint: sm.ImageFootprint :
        sequence of points (ra, dec) or an `ImageFootprint` object
    skymap: sc.SkyMap :
        sky map instance; defaults to global SKYMAP (Default value = None)

    Returns
    -------
    indices of projection regions in the sky map that intersect the given footprint
    """

    if skymap is None:
        skymap = sc.SKYMAP

    # find the closest projection regions to the image center
    _, nearby_projregion_indices = skymap.projection_regions_kdtree.query(
        footprint.vectorpoint_center,
        k=footprint.possibly_intersecting_projregions,
        distance_upper_bound=footprint.possible_intersecting_projregion_distance * 1.1,
    )

    intersecting_projregion_indices = []
    for projregion_index in nearby_projregion_indices:
        if projregion_index < skymap.projection_regions_kdtree.n:
            projregion = sc.ProjectionRegion(projregion_index, skymap=skymap)
            if footprint.polygon.intersects_poly(projregion.polygon):
                intersecting_projregion_indices.append(projregion_index)

    return intersecting_projregion_indices


def veccoords_to_tangent_plane(
    vertices: list[tuple[float, float, float]],
    tangent_vectorpoint: tuple[float, float, float],
) -> NDArray[float]:
    """Convert the spherical geometry vectors to tangent plane coordinates
    in arcseconds. This algorithm is not precise, but should be good
    enough for now (and besides, the goal here is visualizaion, not
    ultra-precision). This also breaks down numerically very near the
    poles.
    """

    # First compute the tangent plane axis vectors.
    x_axis = sgv.normalize_vector(np.cross([0, 0, 1], tangent_vectorpoint))
    y_axis = sgv.normalize_vector(
        np.array([0, 0, 1])
        - np.array(tangent_vectorpoint)
        * np.dot(np.array([0, 0, 1]), np.array(tangent_vectorpoint))
    )
    avertices = np.vstack(vertices).T
    x_coords = np.dot(x_axis, avertices) * RAD_TO_ARCSEC
    y_coords = np.dot(y_axis, avertices) * RAD_TO_ARCSEC
    return np.stack([x_coords, y_coords], axis=1)


def plot_field(corners: NDArray[float], id: str = "", fill=None, color=None, axis=None):
    if axis is None:
        axis = plt
    axis.fill(corners[:, 0], corners[:, 1], color=fill, edgecolor=color)


def plot_projregion(
    projregion: sc.ProjectionRegion, color=None, label: bool = True, axis=None
):
    if axis is None:
        axis = plt

    tangent_vectorpoint = sgv.normalize_vector(
        sgv.lonlat_to_vector(*projregion.radec_tangent)
    )
    corners = projregion.vectorpoint_corners
    corners = np.concat([corners, corners[0, :].reshape((1, 3))], axis=0)
    corners_tangentplane = veccoords_to_tangent_plane(
        corners,
        tangent_vectorpoint,
    )

    axis.plot(corners_tangentplane[:, 0], corners_tangentplane[:, 1], color=color)

    if label:
        center = np.mean(corners_tangentplane[:-1], axis=0)
        axis.annotate(
            f"proj{projregion.index}",
            center,
            va="center",
            ha="center",
            size=10,
            color=color,
        )


def plot_skycell(
    skycell: sc.SkyCell,
    tangent_vectorpoint: tuple[float, float, float],
    color=None,
    label: str | None = None,
    annotation: str | None = None,
    axis=None,
):
    if axis is None:
        axis = plt

    corners = skycell.vectorpoint_corners
    corners = np.concat([corners, corners[0, :].reshape((1, 3))], axis=0)
    corners_tangentplane = veccoords_to_tangent_plane(
        corners,
        tangent_vectorpoint,
    )

    axis.plot(
        corners_tangentplane[:, 0], corners_tangentplane[:, 1], color=color, label=label
    )

    if annotation:
        center = np.mean(corners_tangentplane[:-1], axis=0)
        axis.annotate(
            annotation, center, va="center", ha="center", size=10, color=color
        )


def plot_image_footprint_and_skycells(
    footprint: list[tuple[float, float]] | sm.ImageFootprint,
    skymap: sc.SkyMap = None,
) -> list[tuple[Axis, tuple[float, float, float]]]:
    """This plots a list of skycell footprints against the image footprint.

    Both the touched skycells as well as nearby skycells are plotted.

    Parameters
    ----------
    footprint : list | sm.ImageFootprint :
        sequence of points (ra, dec) or an `ImageFootprint` object
    skymap : sc.SkyMap :
        sky map instance; defaults to global SKYMAP (Default value = None)
    """

    if not isinstance(footprint, sm.ImageFootprint):
        footprint = sm.ImageFootprint(footprint)

    if skymap is None:
        skymap = sc.SKYMAP

    # plot each intersecting projection region's intersection onto a tangent plane
    intersecting_projregion_indices = find_intersecting_projregions(
        footprint, skymap=skymap
    )
    axes = []
    for projregion_index in intersecting_projregion_indices:
        figure = plt.figure()
        figure.gca().invert_xaxis()
        figure.suptitle(f"projection region {projregion_index}")
        axis = figure.subplots(1, 1)
        axis.plot(0, 0, "+", markersize=10)

        projregion = sc.ProjectionRegion(projregion_index, skymap=skymap)

        tangent_vectorpoint = sgv.normalize_vector(
            sgv.lonlat_to_vector(*projregion.radec_tangent)
        )
        image_corners_tangentplane = veccoords_to_tangent_plane(
            footprint.vectorpoint_vertices,
            tangent_vectorpoint,
        )
        plot_field(image_corners_tangentplane, fill="lightgrey", color="black")

        plot_projregion(projregion, color="lightgrey")

        for skycell_index in projregion.skycell_indices:
            skycell = sc.SkyCell(skycell_index, skymap=skymap)
            plot_skycell(
                skycell,
                tangent_vectorpoint,
                color="darkgrey",
            )

        _, nearby_skycell_indices = projregion.skycells_kdtree.query(
            footprint.vectorpoint_center,
            k=footprint.possibly_intersecting_skycells,
            distance_upper_bound=footprint.length + sc.SkyCell.length,
        )

        for skycell_index in nearby_skycell_indices:
            skycell = sc.SkyCell(
                projregion.data["skycell_start"] + skycell_index, skymap=skymap
            )
            plot_skycell(
                skycell,
                tangent_vectorpoint,
                color="red",
            )

            if footprint.polygon.intersects_poly(skycell.polygon):
                plot_skycell(
                    skycell, tangent_vectorpoint, color="blue", annotation=skycell.name
                )

        axis.set_xlabel("Offset from nearest tangent point in arcsec")
        axis.set_ylabel("Offset from nearest tangent point in arcsec")

        axis.set_title(f"tangent point radec {np.array(projregion.radec_tangent)}")

        axes.append((axis, tangent_vectorpoint))

    return axes



from functools import reduce

import numpy as np
import pytest
from asdf import AsdfFile
from astropy import coordinates as coord
from astropy import units as u
from astropy.modeling import models
from astropy.table import Table
from astropy.time import Time
from gwcs import WCS
from gwcs import coordinate_frames as cf
from roman_datamodels import datamodels

from romancal.assign_wcs.utils import add_s_region
from romancal.datamodels import ModelLibrary
from romancal.regtest import util
from romancal.resample import ResampleStep
from romancal.resample.l3_wcs import l3wcsinfo_to_wcs


def create_wcs_object_without_distortion(fiducial_world, pscale, shape, **kwargs):
    """
    Create a simple WCS object without either distortion or rotation.

    Parameters
    ----------
    fiducial_world : tuple
        A pair of values corresponding to the fiducial's world coordinate.
    pscale : tuple
        A pair of values corresponding to the pixel scale in each axis.
    shape : tuple
        A pair of values specifying the dimensions of the WCS object.

    Returns
    -------
    gwcs.WCS
        A gwcs.WCS object.
    """
    # components of the model
    shift = models.Shift() & models.Shift()
    affine = models.AffineTransformation2D(
        matrix=[[1, 0], [0, 1]], translation=[0, 0], name="pc_rotation_matrix"
    )
    scale = models.Scale(pscale[0]) & models.Scale(pscale[1])
    tan = models.Pix2Sky_TAN()
    celestial_rotation = models.RotateNative2Celestial(
        fiducial_world[0],
        fiducial_world[1],
        180,
    )

    # transforms between frames
    # detector -> sky
    det2sky = shift | affine | scale | tan | celestial_rotation
    det2sky.name = "linear_transform"

    # frames
    detector_frame = cf.Frame2D(
        name="detector",
        axes_order=(0, 1),
        axes_names=("x", "y"),
        unit=(u.pix, u.pix),
    )
    sky_frame = cf.CelestialFrame(
        reference_frame=coord.FK5(), name="fk5", unit=(u.deg, u.deg)
    )

    pipeline = [
        (detector_frame, det2sky),
        (sky_frame, None),
    ]

    wcs_obj = WCS(pipeline)

    wcs_obj.bounding_box = kwargs.get(
        "bounding_box",
        (
            (-0.5, shape[-1] - 0.5),
            (-0.5, shape[-2] - 0.5),
        ),
    )

    wcs_obj.pixel_shape = shape[::-1]
    wcs_obj.array_shape = shape

    return wcs_obj


@pytest.fixture
def asdf_wcs_file():
    def _create_asdf_wcs_file(tmp_path):
        file_path = tmp_path / "wcs.asdf"
        wcs_data = create_wcs_object_without_distortion(
            (10, 0),
            (0.000031, 0.000031),
            (100, 100),
        )
        wcs = {"wcs": wcs_data}
        with AsdfFile(wcs) as af:
            af.write_to(file_path)
        return str(file_path)

    return _create_asdf_wcs_file


def test_load_custom_wcs_no_file():
    step = ResampleStep()
    result = step._load_custom_wcs(None, (512, 512))
    assert result is None


def test_load_custom_wcs_missing_output_shape(asdf_wcs_file):
    with pytest.raises(ValueError):
        step = ResampleStep()
        step._load_custom_wcs(asdf_wcs_file, None)


def test_load_custom_wcs_invalid_file(tmp_path):
    step = ResampleStep()
    invalid_file = tmp_path / "invalid.asdf"
    with open(invalid_file, "w") as f:
        f.write("invalid asdf file")

    with pytest.raises(ValueError):
        step._load_custom_wcs(str(invalid_file), (512, 512))


def test_load_custom_wcs_asdf_without_wcs_attribute(tmp_path):
    step = ResampleStep()
    file_path = tmp_path / "asdf_file.asdf"
    wcs = {}
    with AsdfFile(wcs) as af:
        af.write_to(file_path)

    with pytest.raises(KeyError):
        step._load_custom_wcs(str(file_path), (100, 100))


def test_wcs_wcsinfo_matches(base_image):
    model = base_image()
    img = datamodels.ImageModel(model)
    add_s_region(img)

    output_model = ResampleStep().run(img)

    wcs_from_wcsinfo = l3wcsinfo_to_wcs(output_model.meta.wcsinfo)
    ra_mad, dec_mad = util.comp_wcs_grids_arcs(output_model.meta.wcs, wcs_from_wcsinfo)
    assert (ra_mad + dec_mad) / 2.0 < 1.0e-5


@pytest.mark.parametrize(
    "good_bits",
    [
        "~DO_NOT_USE+NON_SCIENCE",
        "~513",
        "~1+512",
        "~1,512",
        "LOW_QE+NONLINEAR",
        "73728",
        "8192+65536",
        "8192,65536",
    ],
)
def test_set_good_bits_in_resample_meta(base_image, good_bits):
    model = base_image()
    model.meta.wcsinfo["vparity"] = -1
    model.meta.wcs.bounding_box = (-0.5, 99.5), (-0.5, 99.5)

    img = datamodels.ImageModel(model)
    add_s_region(img)

    img.data *= img.meta.photometry.conversion_megajanskys / img.data

    step = ResampleStep(good_bits=good_bits)
    res = step.run(img)

    assert res.meta.resample.good_bits == good_bits


def test_individual_image_meta(base_image):
    """Test that the individual_image_meta is being populated"""
    models = [base_image() for _ in range(2)]
    # add a None value
    models[0].meta.photometry.pixel_area = None
    models[1].meta.photometry.pixel_area = 1

    # add "extra" stuff (similar to "wcs_fit_results")
    # this will end up in the "basic" table
    models[1].meta["extra"] = {}
    models[1].meta.extra = {"a": 1}

    # remove "background" from one model
    del models[0].meta["background"]

    input_models = ModelLibrary(models)
    output_model = ResampleStep().run(input_models)

    # Check that blended table doesn't make model invalid
    output_model.validate()

    # Assert sizes are expected
    n_inputs = len(input_models)
    for value in output_model.meta.individual_image_meta.values():
        assert type(value) is Table
        assert len(value) == n_inputs

    # Assert spot check on filename, which is different for each mock input
    basic_table = output_model.meta.individual_image_meta.basic
    with input_models:
        for idx, input in enumerate(input_models):
            assert input.meta.filename == basic_table["filename"][idx]
            input_models.shelve(input, index=idx)

    output_tables = output_model.meta.individual_image_meta

    assert "background" in output_tables

    assert "photometry" in output_tables
    assert np.isnan(output_tables["photometry"]["pixel_area"][0])
    assert output_tables["photometry"]["pixel_area"][1] == 1

    assert "extra" not in output_tables
    assert "extra" not in output_tables["basic"].colnames


@pytest.mark.parametrize(
    "meta_overrides, expected",
    [
        (  # 2 exposures, share visit, etc
            (
                {
                    "meta.observation.visit": 1,
                    "meta.observation.pass": 1,
                    "meta.observation.segment": 1,
                    "meta.observation.program": 1,
                    "meta.instrument.optical_element": "F158",
                    "meta.instrument.name": "WFI",
                },
                {
                    "meta.observation.visit": 1,
                    "meta.observation.pass": 1,
                    "meta.observation.segment": 1,
                    "meta.observation.program": 1,
                    "meta.instrument.optical_element": "F158",
                    "meta.instrument.name": "WFI",
                },
            ),
            {
                "meta.observation.visit": 1,
                "meta.observation.pass": 1,
                "meta.observation.segment": 1,
                "meta.instrument.optical_element": "F158",
                "meta.instrument.name": "WFI",
            },
        ),
        (  # 2 exposures, different metadata
            (
                {
                    "meta.observation.visit": 1,
                    "meta.observation.pass": 1,
                    "meta.observation.segment": 1,
                    "meta.observation.program": 1,
                    "meta.instrument.optical_element": "F158",
                    "meta.instrument.name": "WFI",
                },
                {
                    "meta.observation.visit": 2,
                    "meta.observation.pass": 2,
                    "meta.observation.segment": 2,
                    "meta.observation.program": 2,
                    "meta.instrument.optical_element": "F062",
                    "meta.instrument.name": "WFI",
                },
            ),
            {
                "meta.observation.visit": None,
                "meta.observation.pass": None,
                "meta.observation.segment": None,
                "meta.instrument.optical_element": None,
                "meta.instrument.name": "WFI",
            },
        ),
    ],
)
def test_populate_mosaic_metadata(base_image, meta_overrides, expected):
    """Test that the mosaic metadata is being populated"""
    models = []
    for i, meta_override in enumerate(meta_overrides):
        model = base_image()

        model.meta.observation.observation_id = i

        model.meta.exposure.start_time = Time(59000 + i, format="mjd")
        model.meta.exposure.end_time = Time(59001 + i, format="mjd")
        model.meta.exposure.exposure_time = 3600 * 24

        for key, value in meta_override.items():
            *parent_keys, child_key = key.split(".")
            setattr(reduce(getattr, parent_keys, model), child_key, value)
        models.append(model)

    input_models = ModelLibrary(models)
    output_model = ResampleStep().run(input_models)

    assert output_model.meta.coadd_info.time_first == models[0].meta.exposure.start_time
    assert output_model.meta.coadd_info.time_last == models[-1].meta.exposure.end_time
    assert output_model.meta.coadd_info.time_mean.mjd == np.mean(
        [m.meta.exposure.start_time.mjd for m in models]
    )

    for key, value in expected.items():
        *path, final = key.split(".")
        obj = output_model
        for sub_path in path:
            obj = obj[sub_path]
        assert obj[final] == value



"""Regression tests for the photom step of the Roman pipeline"""

import math

import pytest
import roman_datamodels as rdm

from romancal.stpipe import RomanStep

from .regtestdata import compare_asdf


@pytest.mark.bigdata
def test_absolute_photometric_calibration(
    rtdata, ignore_asdf_paths, resource_tracker, request, dms_logger
):
    """DMS140 Test: Testing application of photometric correction using
    CRDS selected photom file."""

    input_data = "r0000101001001001001_0001_wfi01_f158_flat.asdf"
    rtdata.get_data(f"WFI/image/{input_data}")
    rtdata.input = input_data

    #  In Wide Field Imaging mode, the DMS shall generate Level 2 science
    # data products with absolute photometry calibrated in the WFI filter
    # used for the exposure.
    dms_logger.info(
        "DMS140 MSG: Testing absolute photometric "
        "calibrated image data. "
        "Success is creation of a Level 2 image file with "
        "CRDS selected photom file applied."
    )

    dms_logger.info(f"DMS140 MSG: Image data file: {rtdata.input.rsplit('/', 1)[1]}")

    # Note: if any of the following tests fail, check for a different
    # photom match from CRDS. Values come from roman_wfi_photom_0034.asdf

    # Test PhotomStep
    output = "r0000101001001001001_0001_wfi01_f158_photom.asdf"
    rtdata.output = output
    args = ["romancal.step.PhotomStep", rtdata.input]
    dms_logger.info(
        "DMS140 MSG: Running photometric conversion step."
        " The first ERROR is expected, due to extra CRDS parameters"
        " not having been implemented yet."
    )
    with resource_tracker.track(log=request):
        RomanStep.from_cmdline(args)

    photom_out = rdm.open(rtdata.output)

    dms_logger.info(
        "DMS140 MSG: Photom step recorded as complete? :"
        f" {photom_out.meta.cal_step.photom == 'COMPLETE'}"
    )
    assert photom_out.meta.cal_step.photom == "COMPLETE"

    convval = 0.73678
    dms_logger.info(
        "DMS140 MSG: Photom megajansky conversion calculated? : "
        + str(
            math.isclose(
                photom_out.meta.photometry.conversion_megajanskys,
                convval,
                abs_tol=0.0001,
            )
        )
    )
    assert math.isclose(
        photom_out.meta.photometry.conversion_megajanskys, convval, abs_tol=0.0001
    )

    dms_logger.info(
        "DMS140 MSG: Pixel area in steradians calculated? : "
        + str(
            math.isclose(
                photom_out.meta.photometry.pixel_area,
                2.8083e-13,
                abs_tol=1.0e-17,
            )
        )
    )
    assert math.isclose(
        photom_out.meta.photometry.pixel_area,
        2.8083e-13,
        abs_tol=1.0e-17,
    )

    uncval = 0.02866405
    dms_logger.info(
        "DMS140 MSG: Photom megajansky conversion uncertainty calculated? : "
        + str(
            math.isclose(
                photom_out.meta.photometry.conversion_megajanskys_uncertainty,
                uncval,
                abs_tol=1.0e-6,
            )
        )
    )
    assert math.isclose(
        photom_out.meta.photometry.conversion_megajanskys_uncertainty,
        uncval,
        abs_tol=1.0e-6,
    )

    rtdata.get_truth(f"truth/WFI/image/{output}")
    diff = compare_asdf(rtdata.output, rtdata.truth, **ignore_asdf_paths)
    dms_logger.info(
        "DMS140 MSG: Was the proper absolute photometry calibrated image data produced?"
        f" : {diff.identical}"
    )
    assert diff.identical, diff.report()
